/*
  Copyright (c) 2008, Arturas Moskvinas. All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License. You should have
  received a copy of the GPL license along with this program; if you
  did not, you can find it at http://www.gnu.org/
*/

#ifndef _GNU_SOURCE 
#define _GNU_SOURCE
#endif
#include "httpd.h"
#include "http_log.h"
#include "http_request.h"
#include "ap_mpm.h"
#include "scoreboard.h"
/*getpid*/
#include <sys/types.h>
#include <unistd.h>
/*kill and killpg*/
#include <signal.h>
/*apr_atoi64 and apr_int64_t*/
#include "apr.h"
#include "apr_strings.h"
#include "unixd.h"
#include <stdlib.h>
#include <math.h>
module AP_MODULE_DECLARE_DATA overload_module;
/*apr < 1.3 compatiblity issue*/
#ifndef APR_FPROT_UWRITE
#define APR_FPROT_UWRITE APR_UWRITE
#endif
static int server_limit = 0;
static int thread_limit = 0;

/*configuration*/
typedef struct overload_cfg {
    apr_int64_t oneminloadavg;
    apr_int64_t fiveminloadavg;
    apr_int64_t fifteenminloadavg;
    apr_int64_t enabled;
    apr_int64_t log_requests;
    apr_int64_t kill_requests;
    apr_int64_t lower_precedence;
    apr_int64_t allow_503_page;
    apr_int64_t log_one;
    apr_int64_t log_one_file;
    apr_int64_t mod_status_enabled;
    char *lock_path;
    apr_int64_t kill_signal;
    apr_int64_t kill_long_requests;
    apr_int64_t long_request_time;
    char *kill_lock_path;
    apr_int64_t kill_one;
} overload_cfg;

/*load avg, just in case if some OS specific will be needed*/
typedef struct ov_loadavg {
        apr_off_t one_minute;
        apr_off_t five_minute;
        apr_off_t fifteen_minute;
} ov_loadavg;


static void* overload_dconf_create(apr_pool_t *p, char * dir)
{
    overload_cfg *cfg;
    cfg = (overload_cfg *) apr_palloc(p, sizeof(overload_cfg));
    cfg->oneminloadavg = -1; /*defaut values (ignore)*/
    cfg->fiveminloadavg = -1;
    cfg->fifteenminloadavg = -1;
    cfg->enabled = -1; /*not enabled by default*/
    cfg->log_requests = -1; /*not enabled by default*/
    cfg->kill_requests = -1;
    cfg->lower_precedence = -1;
    cfg->allow_503_page = -1;
    cfg->log_one = -1;
    cfg->log_one_file = -1;
    cfg->lock_path = "/tmp/mod_overload.lock";
    cfg->kill_signal = -1;
    cfg->kill_long_requests = -1;
    cfg->long_request_time = -1;
    cfg->lock_path = "/tmp/mod_overload_kill.lock";
    cfg->kill_one = -1;
    if (ap_find_linked_module("mod_status.c")==NULL)
    {
	/*there is a flaw... if mod_status.c is loaded after mod_timedout, then ap_find_linked_module doesn't see mod_status.c*/
	/*we could check is mod_status.c is loaded on each request, but algorithm for searching module is linear, I think it is a bit to expensive on each request...*/
	cfg->mod_status_enabled = 0;
    }
    else
    {
	cfg->mod_status_enabled = 1;
    }
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
    return (void *)cfg;
}

static void *overload_dconf_merge(apr_pool_t *p, void *conf1, void *conf2)
{
    /*values are inherited from parent if it is not set in current context*/
    overload_cfg *cfg;
    cfg = (overload_cfg *) apr_palloc(p, sizeof(overload_cfg));
    overload_cfg *over_conf1 = (overload_cfg *) conf1;
    overload_cfg *over_conf2 = (overload_cfg *) conf2; 
    cfg->oneminloadavg = over_conf2->oneminloadavg!=-1 ? over_conf2->oneminloadavg : over_conf1->oneminloadavg;
    cfg->fiveminloadavg = over_conf2->fiveminloadavg!=-1 ? over_conf2->fiveminloadavg : over_conf1->fiveminloadavg;
    cfg->fifteenminloadavg = over_conf2->fifteenminloadavg!=-1 ? over_conf2->fifteenminloadavg : over_conf1->fifteenminloadavg;
    cfg->enabled = over_conf2->enabled!=-1 ? over_conf2->enabled : over_conf1->enabled;
    cfg->log_requests = over_conf2->log_requests!=-1 ? over_conf2->log_requests : over_conf1->log_requests;
    cfg->kill_requests = over_conf2->kill_requests!=-1 ? over_conf2->kill_requests : over_conf1->kill_requests;
    cfg->lower_precedence = over_conf2->lower_precedence!=-1 ? over_conf2->lower_precedence : over_conf1->lower_precedence;
    cfg->allow_503_page = over_conf2->allow_503_page!=-1 ? over_conf2->allow_503_page : over_conf1->allow_503_page;
    cfg->kill_signal = over_conf2->kill_signal!=-1 ? over_conf2->kill_signal : over_conf1->kill_signal;
    cfg->kill_long_requests = over_conf2->kill_long_requests!=-1 ? over_conf2->kill_long_requests : over_conf1->kill_long_requests;
    cfg->long_request_time = over_conf2->long_request_time!=-1 ? over_conf2->long_request_time : over_conf1->long_request_time;
    cfg->mod_status_enabled = over_conf2->mod_status_enabled!=-1 ? over_conf2->mod_status_enabled : over_conf1->mod_status_enabled;
    return (void *)cfg;
}

static const char* set1min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->oneminloadavg = apr_atoi64(arg1);
    return NULL;
}

static const char* set5min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->fiveminloadavg = apr_atoi64(arg1);
    return NULL;
}

static const char* set15min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->fifteenminloadavg = apr_atoi64(arg1);
    return NULL;
}

static const char* setenableflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->enabled = arg1;
    return NULL;
}

static const char* setlogflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->log_requests = arg1;
    return NULL;
}

static const char* setprecedenceflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->lower_precedence = arg1;
    return NULL;
}

static const char* setkillflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_requests = arg1;
    return NULL;
}

static const char* set503flag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->allow_503_page = arg1;
    return NULL;
}

static const char* setlogoneflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->log_one = arg1;
    return NULL;
}

static const char* setlogonefile(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->log_one_file = arg1;
    return NULL;
}


static const char* setlockpath(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->lock_path = (char *)arg1;
    return NULL;
}

static const char* setkillsignal(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_signal = apr_atoi64(arg1);
    return NULL;
}

static const char* setkilllongrequest(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_long_requests = arg1;
    return NULL;
}

static const char* setlongrequesttime(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->long_request_time = apr_atoi64(arg1);
    return NULL;
}

static const char* setkilloneflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_one = arg1;
    return NULL;
}

static const char* setkilllockpath(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_lock_path = (char *)arg1;
    return NULL;
}


/*Log currently running process */
static void log_requests(request_rec *r, server_rec *s)
{
    worker_score *w_record;
    int i, j;
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload: Currently running requests:");
    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {
    	    w_record = ap_get_scoreboard_worker(i,j);
	    if ((w_record->status==SERVER_READY) || (w_record->status==SERVER_DEAD)) /*not running workers*/
	    {
	        continue;
	    }
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload[%li]: Hostname:%s, uri: %s, client: %s", r->connection->id, w_record->vhost, w_record->request, w_record->client);
	}
    }
}

/*Kill all currently running processes, including the one which started killing process, not elegant solution*/
static void kill_requests(request_rec *r,  apr_int64_t kill_signal, server_rec *s)
{
    /*For sake of not changing behavior it is allowed to kill request which started kill function, there is another feature see OverloadKillLongRunningRequests */
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload: Killing all running requests");
    kill(0, kill_signal);
}

static void kill_long_running_requests(request_rec *r, apr_int64_t kill_signal, apr_int64_t max_request_time, server_rec *s)
{
    int i, j;
    worker_score *w_record;
    pid_t kill_pid;
    for (i=0; i < server_limit; i++) /*daemons*/
    {
        for (j=0; j < thread_limit; j++)/*threads*/
        {
            w_record = ap_get_scoreboard_worker(i,j);
	    if ((w_record->status==SERVER_BUSY_WRITE) && (apr_time_sec(apr_time_now()-w_record->start_time)>max_request_time))/*what about BUSY_READ?*/
	    {/*not prepared for worker MPM, must be made separate test, because there is separate record for thread ID...*/
		kill_pid=ap_get_scoreboard_process(i)->pid;
		if (kill_pid==0)/*should not happen ever...*/
		{
		    ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "mod_overload: Tried to kill whole apache process tree");
		}
		else if (kill_pid==getpid())
		{
		    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload: Tried to kill itself!");
		}
		else
		{
		    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload: Killing request with server PID: %i", kill_pid);
		    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, "mod_overload[%li]: This server was serving this request: hostname:%s, uri: %s, client: %s, duration: %li", r->connection->id, w_record->vhost, w_record->request, w_record->client, apr_time_sec(apr_time_now()-w_record->start_time));
		    kill(kill_pid, kill_signal);
		}
	    }
	}
    }
}

static apr_status_t lock_file(apr_file_t **file, const char *path, apr_pool_t *pool, server_rec *s)
{
    apr_status_t status;
    char buffer[256];
    if (NULL==path)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Bad path argument passed");
	return APR_BADARG;
    }
    status = apr_file_open(file, path, APR_CREATE | APR_WRITE , APR_FPROT_UWRITE, pool);
    if (APR_SUCCESS!=status)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: File open failed, reason %s", apr_strerror(status, buffer, 255));	
	return status;
    }
    status = apr_file_lock(*file, APR_FLOCK_EXCLUSIVE | APR_FLOCK_NONBLOCK);
    if (APR_SUCCESS!=status)
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: File lock failed, reason %s", apr_strerror(status, buffer, 255));
	apr_file_close(*file);
    }
    return status;
}

static apr_status_t unlock_file(apr_file_t *file, server_rec *s)
{
    apr_status_t status;
    char buffer[256];
    status = apr_file_unlock(file);
    if (status!=APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: Failed to unlock file, reason %s", apr_strerror(status, buffer, 255));	
    }
    return apr_file_close(file);
}

static void mod_overload_method_handler_helper(request_rec *r, overload_cfg *cfg, server_rec *s)
{
    apr_file_t *file; /*for lock file*/
    if (r==NULL)
    {
	return;
    }
/*mod_status must be loaded and extendedStatus enable, without it, there won't be any uri's etc.*/
    if ((cfg->mod_status_enabled>0) && (ap_extended_status>0) && (cfg->log_requests>0) && ((cfg->log_one<=0) || (lock_file(&file, cfg->lock_path, r->pool, s)==APR_SUCCESS)))
    {
        log_requests(r, s);
	if (cfg->log_one>0)
	{
	    unlock_file(file, s);
	}
    }
    else if ((cfg->mod_status_enabled<1) && (cfg->log_requests>0))
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "mod_overload: mod_status.c is not loaded, but request for logging is on, please load mod_status.c");
    }
    else if ((cfg->mod_status_enabled>0) && (ap_extended_status<1) && (cfg->log_requests>0))
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "mod_overload: mod_status.c is loaded, ExtendedStatus is off, but request for logging is on, please enable extendedStatus");
    }
    if (cfg->kill_requests>0)
    {
        kill_requests(r, cfg->kill_signal==-1?9:cfg->kill_signal, s);
    }
    if ((cfg->mod_status_enabled>0) && (ap_extended_status>0) && (cfg->kill_long_requests>0) && ((cfg->kill_one<=0) || (lock_file(&file, cfg->kill_lock_path, r->pool, s)==APR_SUCCESS)))
    {
        if (cfg->long_request_time==-1)
        {
    	    ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,"mod_overload: Kill long request setting is on, but long request time is not set!");		
	}
	else
	{
	    kill_long_running_requests(r, cfg->kill_signal==-1?9:cfg->kill_signal, cfg->long_request_time, s);
	}
	if (cfg->kill_one>0)
	{
	    unlock_file(file, s);
	}
    }
    else if ((cfg->mod_status_enabled<1) && (cfg->kill_long_requests>0))
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "mod_overload: mod_status.c is not loaded, but request killing long request is on, please load mod_status.c");
    }
    else if ((cfg->mod_status_enabled>0) && (ap_extended_status<1) && (cfg->kill_long_requests>0))
    {
	ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "mod_overload: mod_status.c is loaded, ExtendedStatus is off, but request killing long request is on, please enable extendedStatus");
    }
}

static int mod_overload_method_handler (request_rec *r)
{
    double ldavg[3];
    server_rec *s;
    ov_loadavg avg;
    overload_cfg *cfg;
    s = r->server;

    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_overload: %s", "Handling request");
    cfg = (overload_cfg*) ap_get_module_config(r->per_dir_config, &overload_module);

    /*no config or not enabled*/
    if ((cfg==NULL) || (cfg->enabled==0))
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Doing nothing, Overload not enabled");
	return DECLINED;
    }
    
    if (cfg->log_one_file==1)
    {/*this will make mod_overload log everything to common apache error log... not individual logs of virtual hosts*/
	s = NULL;
    }    

    /*check if it is internal process which is looking for 503 error page, if it is, let it through, or else we won't have custom 504 pages*/
    if ((cfg->allow_503_page==1) && (r->status==HTTP_SERVICE_UNAVAILABLE))
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Request status is 503, then proceed (to display ErrorDocument)");
	return DECLINED;
    }

    avg.one_minute = -1;
    avg.five_minute = -1;
    avg.fifteen_minute = -1;
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_overload: %s", "Getting load averages (OS specific functions)");

    if (getloadavg(ldavg, 3)==-1)
    {
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,"mod_overload: %s", "Unable to get load average");
	return DECLINED;
    }

    avg.one_minute = floor(ldavg[0]);
    avg.five_minute = floor(ldavg[1]);
    avg.fifteen_minute = floor(ldavg[2]);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: Average load: 1min %li, 5min %li, 15min %li", avg.one_minute, avg.five_minute, avg.fifteen_minute);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: Configuration load: 1min %li, 5min %li, 15min %li", cfg->oneminloadavg, cfg->fiveminloadavg, cfg->fifteenminloadavg);

    if ((cfg->oneminloadavg!=-1) && (cfg->oneminloadavg<avg.one_minute))
    {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 1 minute load average, current load: %li, max load allowed: %li", avg.one_minute, cfg->oneminloadavg);
	mod_overload_method_handler_helper(r, cfg, s);
	return HTTP_SERVICE_UNAVAILABLE;
    }

    if ((cfg->fiveminloadavg!=-1) && (cfg->fiveminloadavg<avg.five_minute))
    {	
	/* if 1min load average less then 5min, then skip stopping, overload is over, else proceed */
	if ((cfg->lower_precedence<1) || ((cfg->lower_precedence>0) && (avg.one_minute>=avg.five_minute))) 
        {
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 5 minute load average, current load: %li, max load allowed: %li", avg.five_minute, cfg->fiveminloadavg);
	    mod_overload_method_handler_helper(r, cfg, s);
	    return HTTP_SERVICE_UNAVAILABLE;
	}
	else
	{
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 1 minute load is lower than 5minute, skipping, 1min load %li, 5min load %li, configured 5min load %li", avg.one_minute, avg.five_minute, cfg->fiveminloadavg);
	}
    }
    if ((cfg->fifteenminloadavg!=-1) && (cfg->fifteenminloadavg<avg.fifteen_minute))
    {
	/* if 5min load average less then 15min, then skip stopping, overload is over, else proceed */
        if ((cfg->lower_precedence<1) || ((cfg->lower_precedence>0) && (avg.five_minute>=avg.fifteen_minute)))
        {
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 15 minute load average, current load: %li, max load allowed: %li", avg.fifteen_minute, cfg->fifteenminloadavg);
	    mod_overload_method_handler_helper(r, cfg, s);
	    return HTTP_SERVICE_UNAVAILABLE;
	}
	else
	{
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 5 minute load is lower than 15 minute, skipping, 5min load %li, 15min load %li, configured 15 min load %li", avg.five_minute, avg.fifteen_minute, cfg->fifteenminloadavg);
	}
    }/*we checked load avg, so it is OK*/
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Not overloaded, returning OK");
    return DECLINED;
}

static void mod_overload_register_hooks (apr_pool_t *p)
{
    ap_hook_handler(mod_overload_method_handler, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec overload_cmds[] = {
AP_INIT_TAKE1(
"OverloadLoadAvg1min",
set1min,
NULL,
RSRC_CONF|ACCESS_CONF,
"1 min load average, when request is stopped"),

AP_INIT_TAKE1(
"OverloadLoadAvg5min",
set5min,
NULL,
RSRC_CONF|ACCESS_CONF,
"5 min load average, when request is stopped"),

AP_INIT_TAKE1(
"OverloadLoadAvg15min",
set15min,
NULL,
RSRC_CONF|ACCESS_CONF,
"15 min load average, when request is stopped"),

AP_INIT_FLAG(
"OverloadEnable",
setenableflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Enable OverloadModule"),

AP_INIT_FLAG(
"OverloadLogRequests",
setlogflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Log Request which where running after Overload detected load avg too high"),

AP_INIT_FLAG(
"OverloadKillAllRequests",
setkillflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Kill all currently running requests (not recommended)"),

AP_INIT_FLAG(
"OverloadAllow503Request",
set503flag,
NULL,
RSRC_CONF|ACCESS_CONF,
"During overload allow any pages which have 503 status code to proceed (allow custom ErrorPage)"),

AP_INIT_FLAG(
"OverloadLowerHasPrecedence",
setprecedenceflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Do not stop request if 1min load is lower than 5 min load (effect after overloading, server sits idling for too long)"),

AP_INIT_FLAG(
"OverloadLogOneAtATime",
setlogoneflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Module should only log one request listing at a time to avoid congestion"),

AP_INIT_FLAG(
"OverloadLogToSingleFile",
setlogonefile,
NULL,
RSRC_CONF|ACCESS_CONF,
"Module should print out all information to one log"),

AP_INIT_TAKE1(
"OverloadLogOneAtATimeLockFile",
setlockpath,
NULL,
RSRC_CONF|ACCESS_CONF,
"Lock file used to implement one log at a time, somewhere where apache can write it... Default: /tmp/mod_overload.lock"),

AP_INIT_TAKE1(
"OverloadRequestKillSignal",
setkillsignal,
NULL,
RSRC_CONF|ACCESS_CONF,
"Kill signal used to kill request (useful if you need core dump..., default - 9 (SIGKILL))"),

AP_INIT_FLAG(
"OverloadKillLongRunningRequests",
setkilllongrequest,
NULL,
RSRC_CONF|ACCESS_CONF,
"Enable long running request killing functionality"),

AP_INIT_TAKE1(
"OverloadLongRunningRequestTime",
setlongrequesttime,
NULL,
RSRC_CONF|ACCESS_CONF,
"The minimum time needed for request to become long running one"),

AP_INIT_TAKE1(
"OverloadKillLongRunningRequestsAtATimeLockFile",
setkilllockpath,
NULL,
RSRC_CONF|ACCESS_CONF,
"Lock file used to implement one killing at a time, somewhere where apache can write it... Default: /tmp/mod_overload_kill.lock"),

AP_INIT_FLAG(
"OverloadKillLongRunningRequestsAtATime",
setkilloneflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Module should only from one request at a time avoid congestion"),

{NULL}
};


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Data declarations.                                                       */
/*                                                                          */
/* Here are the static cells and structure declarations private to our      */
/* module.                                                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/
module AP_MODULE_DECLARE_DATA overload_module =
{
    STANDARD20_MODULE_STUFF,
    overload_dconf_create,    /* per-directory config creator */
    overload_dconf_merge,     /* rules are the same as per dir... */
    NULL,                     /* server config creator */
    NULL,                     /* server config merger */
    overload_cmds,              	/* command table */
    mod_overload_register_hooks,       /* callback for registering hooks */
};
